#ifdef __cplusplus
extern "C" {
#endif

enum Keys
{
    K_A = 0x01,
    K_B = 0x02,
    K_SE= 0x04,
    K_ST= 0x08,
    K_U = 0x10,
    K_D = 0x20,
    K_L = 0x40,
    K_R = 0x80
};

void BisqBotMonitorCPU(unsigned short PC, unsigned char A, unsigned char X, unsigned char Y, unsigned char S, unsigned char P);

void BisqBotReset();
unsigned char BisqBotInput(unsigned FrameNo);

enum BisqBotInstruction
{
    BisqBotContinue,
    BisqBotRestartStep,
    BisqBotRestart,
    BisqBotNextStep,
    BisqBotAccept
};

enum BisqBotInstruction BisqBotTick(unsigned FrameNo, unsigned Step);

unsigned long long BisqBotEvaluate(unsigned FrameNo);


void BisqBotDisplaySeqSample();


#ifdef BISQBOT_INTERNAL
class BisqBotStateBase
{
public:
    virtual ~BisqBotStateBase() { }
};
const BisqBotStateBase* BisqBotNewSaveState();
void BisqBotLoadState(const BisqBotStateBase* b);

class BisqBotVar
{
public:
    void Set(unsigned char value) { values.push_back(value); }
    void Clear() { values.clear(); }
    bool Has(unsigned char v, unsigned min_count=1) const
    {
        for(unsigned a=0; min_count > 0 && a < values.size(); ++a)
            if(values[a] == v) --min_count;
        return min_count == 0;
    }
    bool LimitedTo(unsigned char a,unsigned char b) const
    {
        for(unsigned ind=0; ind<values.size(); ++ind)
            if(values[ind] != a && values[ind] != b) return false;
        return true;
    }
    bool LimitedTo(unsigned char a) const
    {
        for(unsigned ind=0; ind<values.size(); ++ind)
            if(values[ind] != a) return false;
        return true;
    }
    unsigned Count() const { return values.size(); }
    unsigned Value() const { if(!Count()) return 0; return values[0]; }
private:
    std::vector<unsigned char> values;
};
#endif

extern int BotFrontDisableVideo;

void BotFrontMsg(const char*fmt, ...);
void BotFrontStart();
void BotFrontEnd();
void BotFrontToggle();
int BotFrontActive();
unsigned BotFrontGetKeymask();
void BotFrontIter();
void BisqBotRAMreport();

#ifdef __cplusplus
}
#endif
