#include "ntsc/nes_ntsc.h"
#include "BlarggNtsc.h"

#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <alloca.h>

extern "C" {
  namespace FCEU {
    #include "../types.h"
    #include "../x6502.h"
    #include "../fceu.h"
    #include "../palette.h"
    #include "../drivers/common/vidblit.h"
  }
  
  using FCEU::uint32;
  using FCEU::uint16;
  using FCEU::uint8;
}

#if NES_NTSC_OUT_DEPTH != 24
# error BlarggNtsc only works with NES_NTSC_OUT_DEPTH of 24
#endif

static nes_ntsc_t* ntsc = 0;
static int burst_phase = 0;
static nes_ntsc_setup_t setup = nes_ntsc_composite;

void BlarggNtscInit()
{
	ntsc = new nes_ntsc_t;
	
    /*
    static unsigned char basepal[64 * 3];
    for(unsigned a=0; a<64; ++a)
    {
        basepal[a*3+0] = FCEU::palo[a].r;
        basepal[a*3+1] = FCEU::palo[a].g;
        basepal[a*3+2] = FCEU::palo[a].b;
    }
    setup.base_palette = basepal;
    */
    
    /*
    setup.artifacts  = 0;
    setup.sharpness  = 0.1;
    setup.resolution = 0;  
    setup.fringing   = 0;
    setup.bleed      = 0;*/

	setup.merge_fields = 0;
	
	nes_ntsc_init( ntsc, &setup );
}
void BlarggNtscDone()
{
    delete ntsc; ntsc = 0;
}

int BlarggGetWidth(int xr)
{
    if(!ntsc) BlarggNtscInit();

    return NES_NTSC_OUT_WIDTH(xr);
}

int BlarggGetHeight(int yr)
{
    return yr * 2;
}

int BlarggEmphasis;

static void ScanlineMerge(
    unsigned char* dest_new,
    unsigned char* dest_decay,
    const unsigned char* source,
    unsigned char* backup,
    int stride)
{
    std::memcpy(dest_new, source, stride);
    
    for(int a=0; a<stride; ++a)
    {
        dest_decay[a] = (backup[a] * 2 + dest_new[a]) / 3;
        backup[a] = dest_new[a];
    }
    /*
    std::memcpy(dest_decay, backup, stride);
    std::memcpy(backup, dest_new, stride);
    */
}

void BlarggBlit8ToHigh(unsigned char* src, unsigned char* dest, int xr, int yr, int pitch)
{
    if(!ntsc) BlarggNtscInit();
    if(!dest || !src) return;
    
    BlarggEmphasis = FCEU::GetNESDeemph() << 6;
    
    /*
    for(int n=xr*yr, a=0; a<n; ++a)
        src[a] = ((a % xr) * 16 / 256) + 16 * ((a / xr) * 4 / yr);
    */
    
    burst_phase ^= 1;
    if(setup.merge_fields) burst_phase = 0;
    
    int out_pitch = 4,
        out_width = NES_NTSC_OUT_WIDTH(xr),
        out_stride = out_pitch * out_width;
    
    unsigned char* dest_tmp = dest;
    if(pitch != 32) dest_tmp = (unsigned char*)std::malloc( yr * 2 * out_stride);
    
    /*
    std::fprintf(stderr, "xr %d, yr %d, src %p, dest %p, out_width %d, pitch %d\n",
        xr,yr, src, dest, out_width, pitch);
    */
    nes_ntsc_blit(ntsc, src, xr, burst_phase, xr, yr, dest_tmp, out_stride);
    
    static unsigned char merged_out[2500 * 256];
    
    /* double the scanlines */
    for(int y=yr; --y >= 0; )
    {
        if(setup.merge_fields)
        {
            std::memcpy(
                dest_tmp + (y*2) * out_stride,
                dest_tmp + (y) * out_stride,
                out_stride);
            std::memcpy(
                dest_tmp + (y*2+1) * out_stride,
                dest_tmp + (y) * out_stride,
                out_stride);
        }
        else
        {
            ScanlineMerge(
                dest_tmp + (y*2+  burst_phase) * out_stride, // new line
                dest_tmp + (y*2+1-burst_phase) * out_stride, // decaying line
                dest_tmp + (y) * out_stride,
                merged_out + y * out_stride,
                out_stride);
        }
    }
    
    switch(pitch)
    {
        case 32:
            // already done
            break;
        case 24:
            FCEU::Blit32to24( (uint32*)dest_tmp, (uint8*)dest, out_width, yr*2, out_width*3);
            break;
        case 16:
        case 15:
        {
            static int
                shiftr[3] = {8-5, 16-6, 24-5},
                shiftl[3] = {0, 5, 11};
            
            FCEU::Blit32to16( (uint32*)dest_tmp, (uint16*)dest, out_width, yr*2, out_width*2,
                shiftr, shiftl);
            break;
        }
    }
    
    if(pitch != 32) std::free(dest_tmp);

    /*
    for(int n=yr*2*out_stride, a=0; a<n; ++a)
        dest[a] = a;
    */
}
